# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import queue
import threading
import time
import types


class FairSemaphore:
    """Semaphore class that notifies in order of request.

    We cannot use a normal Semaphore because it doesn't give any ordering,
    which could lead to a request starving. Instead, handle them in the
    order we receive them.

    :param int concurrency:
        How many concurrent threads can have the semaphore at once.
    :param float rate_delay:
        How long to wait between the start of each thread receiving the
        semaphore.
    """

    def __init__(self, concurrency: int | None, rate_delay: float):
        self._lock = threading.Lock()
        self._concurrency = concurrency
        if concurrency:
            self._count = 0
            self._queue: queue.Queue[threading.Event] = queue.Queue()

        self._rate_delay = rate_delay
        self._rate_last_ts = time.time()

    def __enter__(self) -> None:
        """Aquire a semaphore."""
        # If concurrency is None, everyone is free to immediately execute.
        if not self._concurrency:
            # NOTE: Rate limiting still applies.This will ultimately impact
            # concurrency a bit due to the mutex.
            with self._lock:
                execution_time = self._advance_timer()
        else:
            execution_time = self._get_ticket()
        return self._wait_for_execution(execution_time)

    def _wait_for_execution(self, execution_time: float) -> None:
        """Wait until the pre-calculated time to run."""
        wait_time = execution_time - time.time()
        if wait_time > 0:
            time.sleep(wait_time)

    def _get_ticket(self) -> float:
        # narrow types - we will only get here if self._concurrency is set
        assert self._concurrency is not None  # nosec B101
        assert self._count is not None  # nosec B101

        ticket = threading.Event()
        with self._lock:
            if self._count <= self._concurrency:
                # We can execute, no need to wait. Take a ticket and
                # move on.
                self._count += 1
                return self._advance_timer()
            else:
                # We need to wait for a ticket before we can execute.
                # Put ourselves in the ticket queue to be woken up
                # when available.
                self._queue.put(ticket)
        ticket.wait()
        with self._lock:
            return self._advance_timer()

    def _advance_timer(self) -> float:
        """Calculate the time when it's ok to run a command again.

        This runs inside of the mutex, serializing the calculation
        of when it's ok to run again and setting _rate_last_ts to that
        new time so that the next thread to calculate when it's safe to
        run starts from the time that the current thread calculated.
        """
        self._rate_last_ts = self._rate_last_ts + self._rate_delay
        return self._rate_last_ts

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        """Release the semaphore."""
        # If concurrency is None, everyone is free to immediately execute
        if not self._concurrency:
            return
        with self._lock:
            # If waiters, wake up the next item in the queue (note
            # we're under the queue lock so the queue won't change
            # under us).
            if self._queue.qsize() > 0:
                ticket = self._queue.get()
                ticket.set()
            else:
                # Nothing else to do, give our ticket back
                self._count -= 1



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from keystoneauth1 import discover
from keystoneauth1.identity.generic import base
from keystoneauth1.identity import v2
from keystoneauth1.identity import v3
from keystoneauth1 import session as ks_session


class Password(base.BaseGenericPlugin):
    """A common user/password authentication plugin.

    :param string username: Username for authentication.
    :param string user_id: User ID for authentication.
    :param string password: Password for authentication.
    :param string user_domain_id: User's domain ID for authentication.
    :param string user_domain_name: User's domain name for authentication.
    """

    def __init__(
        self,
        auth_url: str,
        username: str | None = None,
        user_id: str | None = None,
        password: str | None = None,
        user_domain_id: str | None = None,
        user_domain_name: str | None = None,
        *,
        tenant_id: str | None = None,
        tenant_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        system_scope: str | None = None,
        trust_id: str | None = None,
        default_domain_id: str | None = None,
        default_domain_name: str | None = None,
        reauthenticate: bool = True,
    ):
        super().__init__(
            auth_url=auth_url,
            tenant_id=tenant_id,
            tenant_name=tenant_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            domain_id=domain_id,
            domain_name=domain_name,
            system_scope=system_scope,
            trust_id=trust_id,
            default_domain_id=default_domain_id,
            default_domain_name=default_domain_name,
            reauthenticate=reauthenticate,
        )

        self._username = username
        self._user_id = user_id
        self._password = password
        self._user_domain_id = user_domain_id
        self._user_domain_name = user_domain_name

    def create_plugin(
        self,
        session: ks_session.Session,
        version: discover._PARSED_VERSION_T,
        url: str,
        raw_status: str | None = None,
    ) -> None | v2.Password | v3.Password:
        if discover.version_match((2,), version):
            if self._user_domain_id or self._user_domain_name:
                # TODO(stephenfin): Shouldn't this be an error?
                return None

            if self._password is None:
                # FIXME(stephenfin): It would be better is password was a
                # non-optional paramter to this plugin but that requires
                # changing the __init__ signature
                raise Exception('password is a required attribute')

            return v2.Password(
                auth_url=url,
                user_id=self._user_id,
                username=self._username,
                password=self._password,
                trust_id=self._trust_id,
                tenant_id=self._project_id,
                tenant_name=self._project_name,
                reauthenticate=self.reauthenticate,
            )

        elif discover.version_match((3,), version):
            if self._password is None:
                # FIXME(stephenfin): It would be better is password was a
                # non-optional paramter to this plugin but that requires
                # changing the __init__ signature
                raise Exception('password is a required attribute')

            u_domain_id = self._user_domain_id or self._default_domain_id
            u_domain_name = self._user_domain_name or self._default_domain_name

            return v3.Password(
                auth_url=url,
                user_id=self._user_id,
                username=self._username,
                user_domain_id=u_domain_id,
                user_domain_name=u_domain_name,
                password=self._password,
                trust_id=self._trust_id,
                system_scope=self._system_scope,
                project_id=self._project_id,
                project_name=self._project_name,
                project_domain_id=self.project_domain_id,
                project_domain_name=self.project_domain_name,
                domain_id=self._domain_id,
                domain_name=self._domain_name,
                reauthenticate=self.reauthenticate,
            )

        return None

    @property
    def user_domain_id(self) -> str | None:
        return self._user_domain_id or self._default_domain_id

    @user_domain_id.setter
    def user_domain_id(self, value: str) -> None:
        self._user_domain_id = value

    @property
    def user_domain_name(self) -> str | None:
        return self._user_domain_name or self._default_domain_name

    @user_domain_name.setter
    def user_domain_name(self, value: str) -> None:
        self._user_domain_name = value

    def get_cache_id_elements(self) -> dict[str, str | None]:
        return {
            'auth_url': self.auth_url,
            'project_id': self._project_id,
            'project_name': self._project_name,
            'project_domain_id': self.project_domain_id,
            'project_domain_name': self.project_domain_name,
            'domain_id': self._domain_id,
            'domain_name': self._domain_name,
            'trust_id': self._trust_id,
            'username': self._username,
            'user_id': self._user_id,
            'password': self._password,
            'user_domain_id': self.user_domain_id,
            'user_domain_name': self.user_domain_name,
        }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import time
from unittest import mock
import urllib
import uuid

from keystoneauth1 import exceptions
from keystoneauth1.identity.v3 import oidc
from keystoneauth1 import session
from keystoneauth1.tests.unit import oidc_fixtures
from keystoneauth1.tests.unit import utils


KEYSTONE_TOKEN_VALUE = uuid.uuid4().hex
BUILTIN_INPUT = "builtins.input"


class BaseOIDCTests:
    def setUp(self):
        super().setUp()
        self.session = session.Session()

        self.AUTH_URL = 'http://keystone:5000/v3'
        self.IDENTITY_PROVIDER = 'bluepages'
        self.PROTOCOL = 'oidc'
        self.USER_NAME = 'oidc_user@example.com'
        self.PROJECT_NAME = 'foo project'
        self.PASSWORD = uuid.uuid4().hex
        self.CLIENT_ID = uuid.uuid4().hex
        self.CLIENT_SECRET = uuid.uuid4().hex
        self.ACCESS_TOKEN = uuid.uuid4().hex
        self.ACCESS_TOKEN_ENDPOINT = 'https://localhost:8020/oidc/token'
        self.FEDERATION_AUTH_URL = '{}/{}'.format(
            self.AUTH_URL,
            'OS-FEDERATION/identity_providers/bluepages/protocols/oidc/auth',
        )
        self.REDIRECT_URL = 'urn:ietf:wg:oauth:2.0:oob'
        self.CODE = '4/M9TNz2G9WVwYxSjx0w9AgA1bOmryJltQvOhQMq0czJs.cnLNVAfqwG'

        self.DISCOVERY_URL = (
            'https://localhost:8020/oidc/.well-known/openid-configuration'
        )
        self.SCOPE = 'openid profile'
        self.GRANT_TYPE = None

    def test_discovery_not_found(self):
        self.requests_mock.get("http://not.found", status_code=404)

        plugin = self.plugin.__class__(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            discovery_endpoint="http://not.found",
        )

        self.assertRaises(
            exceptions.http.NotFound,
            plugin._get_discovery_document,
            self.session,
        )

    def test_no_discovery(self):
        plugin = self.plugin.__class__(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
        )
        self.assertEqual(
            self.ACCESS_TOKEN_ENDPOINT, plugin.access_token_endpoint
        )

    def test_load_discovery(self):
        self.requests_mock.get(
            self.DISCOVERY_URL, json=oidc_fixtures.DISCOVERY_DOCUMENT
        )

        plugin = self.plugin.__class__(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            discovery_endpoint=self.DISCOVERY_URL,
        )
        self.assertEqual(
            oidc_fixtures.DISCOVERY_DOCUMENT["token_endpoint"],
            plugin._get_access_token_endpoint(self.session),
        )

    def test_no_access_token_endpoint(self):
        plugin = self.plugin.__class__(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
        )

        self.assertRaises(
            exceptions.OidcAccessTokenEndpointNotFound,
            plugin._get_access_token_endpoint,
            self.session,
        )

    def test_invalid_discovery_document(self):
        self.requests_mock.get(self.DISCOVERY_URL, json={})

        plugin = self.plugin.__class__(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            discovery_endpoint=self.DISCOVERY_URL,
        )

        self.assertRaises(
            exceptions.InvalidOidcDiscoveryDocument,
            plugin._get_discovery_document,
            self.session,
        )

    def test_load_discovery_override_by_endpoints(self):
        self.requests_mock.get(
            self.DISCOVERY_URL, json=oidc_fixtures.DISCOVERY_DOCUMENT
        )

        access_token_endpoint = uuid.uuid4().hex
        plugin = self.plugin.__class__(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            discovery_endpoint=self.DISCOVERY_URL,
            access_token_endpoint=access_token_endpoint,
        )
        self.assertEqual(
            access_token_endpoint,
            plugin._get_access_token_endpoint(self.session),
        )

    def test_wrong_grant_type(self):
        self.requests_mock.get(
            self.DISCOVERY_URL, json={"grant_types_supported": ["foo", "bar"]}
        )

        plugin = self.plugin.__class__(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            discovery_endpoint=self.DISCOVERY_URL,
        )

        self.assertRaises(
            exceptions.OidcPluginNotSupported,
            plugin.get_unscoped_auth_ref,
            self.session,
        )


class OIDCClientCredentialsTests(BaseOIDCTests, utils.TestCase):
    def setUp(self):
        super().setUp()

        self.GRANT_TYPE = 'client_credentials'

        self.plugin = oidc.OidcClientCredentials(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
            project_name=self.PROJECT_NAME,
        )

    def test_initial_call_to_get_access_token(self):
        """Test initial call, expect JSON access token."""
        # Mock the output that creates the access token
        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP,
        )

        # Prep all the values and send the request
        scope = 'profile email'
        payload = {'grant_type': self.GRANT_TYPE, 'scope': scope}
        self.plugin._get_access_token(self.session, payload)

        # Verify the request matches the expected structure
        last_req = self.requests_mock.last_request
        self.assertEqual(self.ACCESS_TOKEN_ENDPOINT, last_req.url)
        self.assertEqual('POST', last_req.method)
        encoded_payload = urllib.parse.urlencode(payload)
        self.assertEqual(encoded_payload, last_req.body)

    def test_second_call_to_protected_url(self):
        """Test subsequent call, expect Keystone token."""
        # Mock the output that creates the keystone token
        self.requests_mock.post(
            self.FEDERATION_AUTH_URL,
            json=oidc_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': KEYSTONE_TOKEN_VALUE},
        )

        res = self.plugin._get_keystone_token(self.session, self.ACCESS_TOKEN)

        # Verify the request matches the expected structure
        self.assertEqual(self.FEDERATION_AUTH_URL, res.request.url)
        self.assertEqual('POST', res.request.method)

        headers = {'Authorization': 'Bearer ' + self.ACCESS_TOKEN}
        self.assertEqual(
            headers['Authorization'], res.request.headers['Authorization']
        )

    def test_end_to_end_workflow(self):
        """Test full OpenID Connect workflow."""
        # Mock the output that creates the access token
        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP,
        )

        # Mock the output that creates the keystone token
        self.requests_mock.post(
            self.FEDERATION_AUTH_URL,
            json=oidc_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': KEYSTONE_TOKEN_VALUE},
        )

        response = self.plugin.get_unscoped_auth_ref(self.session)
        self.assertEqual(KEYSTONE_TOKEN_VALUE, response.auth_token)


class OIDCPasswordTests(BaseOIDCTests, utils.TestCase):
    def setUp(self):
        super().setUp()

        self.GRANT_TYPE = 'password'

        self.plugin = oidc.OidcPassword(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
            project_name=self.PROJECT_NAME,
            username=self.USER_NAME,
            password=self.PASSWORD,
        )

    def test_initial_call_to_get_access_token(self):
        """Test initial call, expect JSON access token."""
        # Mock the output that creates the access token
        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP,
        )

        # Prep all the values and send the request
        grant_type = 'password'
        scope = 'profile email'
        payload = {
            'grant_type': grant_type,
            'username': self.USER_NAME,
            'password': self.PASSWORD,
            'scope': scope,
        }
        self.plugin._get_access_token(self.session, payload)

        # Verify the request matches the expected structure
        last_req = self.requests_mock.last_request
        self.assertEqual(self.ACCESS_TOKEN_ENDPOINT, last_req.url)
        self.assertEqual('POST', last_req.method)
        encoded_payload = urllib.parse.urlencode(payload)
        self.assertEqual(encoded_payload, last_req.body)

    def test_second_call_to_protected_url(self):
        """Test subsequent call, expect Keystone token."""
        # Mock the output that creates the keystone token
        self.requests_mock.post(
            self.FEDERATION_AUTH_URL,
            json=oidc_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': KEYSTONE_TOKEN_VALUE},
        )

        res = self.plugin._get_keystone_token(self.session, self.ACCESS_TOKEN)

        # Verify the request matches the expected structure
        self.assertEqual(self.FEDERATION_AUTH_URL, res.request.url)
        self.assertEqual('POST', res.request.method)

        headers = {'Authorization': 'Bearer ' + self.ACCESS_TOKEN}
        self.assertEqual(
            headers['Authorization'], res.request.headers['Authorization']
        )

    def test_end_to_end_workflow(self):
        """Test full OpenID Connect workflow."""
        # Mock the output that creates the access token
        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP,
        )

        # Mock the output that creates the keystone token
        self.requests_mock.post(
            self.FEDERATION_AUTH_URL,
            json=oidc_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': KEYSTONE_TOKEN_VALUE},
        )

        response = self.plugin.get_unscoped_auth_ref(self.session)
        self.assertEqual(KEYSTONE_TOKEN_VALUE, response.auth_token)

    @mock.patch(BUILTIN_INPUT)
    def test_otp_while_generating_the_access_token(self, mock_input):
        """Test the configured otp in the access token flow."""
        self.plugin.idp_otp_key = "totp"
        otp = 123
        new_otp = 111
        access_token = oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP[
            "access_token"
        ]

        mock_input.return_value = otp

        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP,
        )

        # Verify if the OTP is set in the session and if the request is OK
        payload = self.plugin.get_payload(self.session)
        resp = self.plugin._get_access_token(self.session, payload)
        self.assertEqual(resp, access_token)
        self.assertEqual(self.session.otp, otp)

        # Verify if the OTP is obtained from the session.
        mock_input.return_value = new_otp
        payload = self.plugin.get_payload(self.session)
        self.assertEqual(payload[self.plugin.idp_otp_key], otp)


class OIDCAuthorizationGrantTests(BaseOIDCTests, utils.TestCase):
    def setUp(self):
        super().setUp()

        self.GRANT_TYPE = 'authorization_code'

        self.plugin = oidc.OidcAuthorizationCode(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
            redirect_uri=self.REDIRECT_URL,
            project_name=self.PROJECT_NAME,
            code=self.CODE,
        )

    def test_initial_call_to_get_access_token(self):
        """Test initial call, expect JSON access token."""
        # Mock the output that creates the access token
        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_AUTH_GRANT_RESP,
        )

        # Prep all the values and send the request
        grant_type = 'authorization_code'
        payload = {
            'grant_type': grant_type,
            'redirect_uri': self.REDIRECT_URL,
            'code': self.CODE,
        }
        self.plugin._get_access_token(self.session, payload)

        # Verify the request matches the expected structure
        last_req = self.requests_mock.last_request
        self.assertEqual(self.ACCESS_TOKEN_ENDPOINT, last_req.url)
        self.assertEqual('POST', last_req.method)
        encoded_payload = urllib.parse.urlencode(payload)
        self.assertEqual(encoded_payload, last_req.body)


# NOTE(aloga): This is a special case, as we do not need all the other openid
# parameters, like client_id, client_secret, access_token_endpoint and so on,
# therefore we do not inherit from the base oidc test class, but from the base
# TestCase
class OIDCTokenTests(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.session = session.Session()

        self.AUTH_URL = 'http://keystone:5000/v3'
        self.IDENTITY_PROVIDER = 'bluepages'
        self.PROTOCOL = 'oidc'
        self.PROJECT_NAME = 'foo project'
        self.ACCESS_TOKEN = uuid.uuid4().hex

        self.FEDERATION_AUTH_URL = '{}/{}'.format(
            self.AUTH_URL,
            'OS-FEDERATION/identity_providers/bluepages/protocols/oidc/auth',
        )

        self.plugin = oidc.OidcAccessToken(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            access_token=self.ACCESS_TOKEN,
            project_name=self.PROJECT_NAME,
        )

    def test_end_to_end_workflow(self):
        """Test full OpenID Connect workflow."""
        # Mock the output that creates the keystone token
        self.requests_mock.post(
            self.FEDERATION_AUTH_URL,
            json=oidc_fixtures.UNSCOPED_TOKEN,
            headers={'X-Subject-Token': KEYSTONE_TOKEN_VALUE},
        )

        response = self.plugin.get_unscoped_auth_ref(self.session)
        self.assertEqual(KEYSTONE_TOKEN_VALUE, response.auth_token)


class OIDCDeviceAuthorizationTest(BaseOIDCTests, utils.TestCase):
    def setUp(self):
        super().setUp()

        self.GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:device_code'
        self.DEVICE_AUTH_ENDPOINT = (
            'https://localhost:8020/oidc/authorize/device'
        )

        self.plugin = oidc.OidcDeviceAuthorization(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            client_secret=self.CLIENT_SECRET,
            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
            device_authorization_endpoint=self.DEVICE_AUTH_ENDPOINT,
        )

    def test_create_device_auth_plugin_without_client_secret(self):
        self.plugin = oidc.OidcDeviceAuthorization(
            self.AUTH_URL,
            self.IDENTITY_PROVIDER,
            self.PROTOCOL,
            client_id=self.CLIENT_ID,
            access_token_endpoint=self.ACCESS_TOKEN_ENDPOINT,
            device_authorization_endpoint=self.DEVICE_AUTH_ENDPOINT,
        )

    def test_get_device_authorization_request(self):
        """Test device authorization request."""
        # Mock the output that creates the device code
        self.requests_mock.post(
            self.DEVICE_AUTH_ENDPOINT, json=oidc_fixtures.DEVICE_CODE_RESP
        )

        result = self.plugin.get_payload(self.session)

        # Verify the request matches the expected structure
        last_req = self.requests_mock.last_request
        self.assertEqual(self.DEVICE_AUTH_ENDPOINT, last_req.url)
        self.assertEqual('POST', last_req.method)
        self.assertEqual(
            self._basic_header(self.CLIENT_ID, self.CLIENT_SECRET),
            last_req.headers.get('Authorization'),
        )
        self.assertEqual(
            oidc_fixtures.DEVICE_CODE_RESP["device_code"],
            result["device_code"],
        )

    def test_get_device_authorization_request_without_client_secret(self):
        """Test device authorization request, without client_secret.

        From RFC 8628 Section 3.1, if Basic authentication is not performed,
        it is required to include the client_id in the request body.
        """
        self.plugin.client_secret = None

        # Mock the output that creates the device code
        self.requests_mock.post(
            self.DEVICE_AUTH_ENDPOINT, json=oidc_fixtures.DEVICE_CODE_RESP
        )
        result = self.plugin.get_payload(self.session)

        payload = {'client_id': self.CLIENT_ID, 'scope': self.SCOPE}
        # Verify the request matches the expected structure
        last_req = self.requests_mock.last_request
        self.assertEqual(self.DEVICE_AUTH_ENDPOINT, last_req.url)
        self.assertEqual('POST', last_req.method)
        self.assertIsNone(last_req.headers.get('Authorization'))
        encoded_payload = urllib.parse.urlencode(payload)
        self.assertEqual(encoded_payload, last_req.body)
        self.assertEqual(
            oidc_fixtures.DEVICE_CODE_RESP["device_code"],
            result["device_code"],
        )

    def test_second_call_to_get_access_token(self):
        """Test Device Access Token Request."""
        self._store_device_authorization_response()
        # Mock the output that creates the access token
        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP,
        )

        # Prep all the values and send the request
        payload = {
            'grant_type': self.GRANT_TYPE,
            'device_code': self.plugin.device_code,
        }
        self.plugin._get_access_token(self.session, payload)

        # Verify the request matches the expected structure
        last_req = self.requests_mock.last_request
        self.assertEqual(self.ACCESS_TOKEN_ENDPOINT, last_req.url)
        self.assertEqual('POST', last_req.method)
        self.assertEqual(
            self._basic_header(self.CLIENT_ID, self.CLIENT_SECRET),
            last_req.headers.get('Authorization'),
        )
        encoded_payload = urllib.parse.urlencode(payload)
        self.assertEqual(encoded_payload, last_req.body)

    def test_second_call_to_get_access_token_without_client_secret(self):
        """Device Access Token Request without client_secret.

        From RFC 8628 Section 3.4, if Basic authentication is not performed,
        it is required to include the client_id in the request body.
        """
        self.plugin.client_secret = None
        self._store_device_authorization_response()

        # Mock the output that creates the access token
        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            json=oidc_fixtures.ACCESS_TOKEN_VIA_PASSWORD_RESP,
        )

        # Prep all the values and send the request
        payload = {
            'grant_type': self.GRANT_TYPE,
            'device_code': self.plugin.device_code,
            'client_id': self.CLIENT_ID,
        }
        self.plugin._get_access_token(self.session, payload)

        # Verify the request matches the expected structure
        last_req = self.requests_mock.last_request
        self.assertEqual(self.ACCESS_TOKEN_ENDPOINT, last_req.url)
        self.assertEqual('POST', last_req.method)
        self.assertIsNone(last_req.headers.get('Authorization'))
        encoded_payload = urllib.parse.urlencode(payload)
        self.assertEqual(encoded_payload, last_req.body)

    @mock.patch('time.sleep')
    def test_second_call_to_get_access_token_authorization_pending(
        self, m_sleep
    ):
        """Test Device Access Token Request."""
        self._store_device_authorization_response()

        self.requests_mock.post(
            self.ACCESS_TOKEN_ENDPOINT,
            status_code=400,
            json={'error': 'authorization_pending'},
            headers={'Content-Type': 'application/json'},
        )
        m_sleep.side_effect = exceptions.oidc.OidcDeviceAuthorizationTimeOut()

        # Prep all the values and send the request
        payload = {
            'grant_type': self.GRANT_TYPE,
            'device_code': self.plugin.device_code,
        }
        self.assertRaises(
            exceptions.oidc.OidcDeviceAuthorizationTimeOut,
            self.plugin._get_access_token,
            self.session,
            payload,
        )

    def _basic_header(self, username, password):
        user_pass = f'{username}:{password}'.encode()
        encoded_credentials = base64.b64encode(user_pass).decode('utf-8')
        return f'Basic {encoded_credentials}'

    def _store_device_authorization_response(self):
        self.plugin.expires_in = int(
            oidc_fixtures.DEVICE_CODE_RESP["expires_in"]
        )
        self.plugin.timeout = time.time() + self.plugin.expires_in
        self.plugin.device_code = oidc_fixtures.DEVICE_CODE_RESP["device_code"]
        self.plugin.interval = int(oidc_fixtures.DEVICE_CODE_RESP["interval"])
        self.plugin.user_code = oidc_fixtures.DEVICE_CODE_RESP["user_code"]
        self.plugin.verification_uri = oidc_fixtures.DEVICE_CODE_RESP[
            "verification_uri"
        ]
        self.plugin.verification_uri_complete = oidc_fixtures.DEVICE_CODE_RESP[
            "verification_uri_complete"
        ]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1 import fixture
from keystoneauth1.tests.unit import utils


class ServiceCatalogTest(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.AUTH_RESPONSE_BODY = fixture.V3Token(
            audit_chain_id=uuid.uuid4().hex
        )
        self.AUTH_RESPONSE_BODY.set_project_scope()

        self.AUTH_RESPONSE_BODY.add_role(name='admin')
        self.AUTH_RESPONSE_BODY.add_role(name='member')

        s = self.AUTH_RESPONSE_BODY.add_service('compute', name='nova')
        s.add_standard_endpoints(
            public='https://compute.north.host/novapi/public',
            internal='https://compute.north.host/novapi/internal',
            admin='https://compute.north.host/novapi/admin',
            region='North',
        )

        s = self.AUTH_RESPONSE_BODY.add_service('object-store', name='swift')
        s.add_standard_endpoints(
            public='http://swift.north.host/swiftapi/public',
            internal='http://swift.north.host/swiftapi/internal',
            admin='http://swift.north.host/swiftapi/admin',
            region='South',
        )

        s = self.AUTH_RESPONSE_BODY.add_service('image', name='glance')
        s.add_standard_endpoints(
            public='http://glance.north.host/glanceapi/public',
            internal='http://glance.north.host/glanceapi/internal',
            admin='http://glance.north.host/glanceapi/admin',
            region='North',
        )

        s.add_standard_endpoints(
            public='http://glance.south.host/glanceapi/public',
            internal='http://glance.south.host/glanceapi/internal',
            admin='http://glance.south.host/glanceapi/admin',
            region='South',
        )

        s = self.AUTH_RESPONSE_BODY.add_service('block-storage', name='cinder')
        s.add_standard_endpoints(
            public='http://cinder.north.host/cinderapi/public',
            internal='http://cinder.north.host/cinderapi/internal',
            admin='http://cinder.north.host/cinderapi/admin',
            region='North',
        )

        s = self.AUTH_RESPONSE_BODY.add_service('volumev2', name='cinder')
        s.add_standard_endpoints(
            public='http://cinder.south.host/cinderapi/public/v2',
            internal='http://cinder.south.host/cinderapi/internal/v2',
            admin='http://cinder.south.host/cinderapi/admin/v2',
            region='South',
        )

        s = self.AUTH_RESPONSE_BODY.add_service('volumev3', name='cinder')
        s.add_standard_endpoints(
            public='http://cinder.south.host/cinderapi/public/v3',
            internal='http://cinder.south.host/cinderapi/internal/v3',
            admin='http://cinder.south.host/cinderapi/admin/v3',
            region='South',
        )

        self.north_endpoints = {
            'public': 'http://glance.north.host/glanceapi/public',
            'internal': 'http://glance.north.host/glanceapi/internal',
            'admin': 'http://glance.north.host/glanceapi/admin',
        }

        self.south_endpoints = {
            'public': 'http://glance.south.host/glanceapi/public',
            'internal': 'http://glance.south.host/glanceapi/internal',
            'admin': 'http://glance.south.host/glanceapi/admin',
        }

    def test_building_a_service_catalog(self):
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog

        self.assertEqual(
            sc.url_for(service_type='compute'),
            "https://compute.north.host/novapi/public",
        )
        self.assertEqual(
            sc.url_for(service_type='compute', interface='internal'),
            "https://compute.north.host/novapi/internal",
        )

        self.assertRaises(
            exceptions.EndpointNotFound,
            sc.url_for,
            region_name='South',
            service_type='compute',
        )

    def test_service_catalog_endpoints(self):
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog

        public_ep = sc.get_endpoints(
            service_type='compute', interface='public'
        )
        self.assertEqual(public_ep['compute'][0]['region'], 'North')
        self.assertEqual(
            public_ep['compute'][0]['url'],
            "https://compute.north.host/novapi/public",
        )

    def test_service_catalog_alias_find_official(self):
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog

        # Tests that we find the block-storage endpoint when we request
        # the volume endpoint.
        public_ep = sc.get_endpoints(
            service_type='volume', interface='public', region_name='North'
        )
        self.assertEqual(public_ep['block-storage'][0]['region'], 'North')
        self.assertEqual(
            public_ep['block-storage'][0]['url'],
            "http://cinder.north.host/cinderapi/public",
        )

    def test_service_catalog_alias_find_exact_match(self):
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog

        # Tests that we find the volumev3 endpoint when we request it.
        public_ep = sc.get_endpoints(
            service_type='volumev3', interface='public'
        )
        self.assertEqual(public_ep['volumev3'][0]['region'], 'South')
        self.assertEqual(
            public_ep['volumev3'][0]['url'],
            "http://cinder.south.host/cinderapi/public/v3",
        )

    def test_service_catalog_alias_find_best_match(self):
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog

        # Tests that we find the volumev3 endpoint when we request
        # block-storage when only volumev2 and volumev3 are present since
        # volumev3 comes first in the list.
        public_ep = sc.get_endpoints(
            service_type='block-storage',
            interface='public',
            region_name='South',
        )
        self.assertEqual(public_ep['volumev3'][0]['region'], 'South')
        self.assertEqual(
            public_ep['volumev3'][0]['url'],
            "http://cinder.south.host/cinderapi/public/v3",
        )

    def test_service_catalog_alias_all_by_name(self):
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog

        # Tests that we find all the cinder endpoints since we request
        # them by name and that no filtering related to aliases happens.
        public_ep = sc.get_endpoints(service_name='cinder', interface='public')
        self.assertEqual(public_ep['volumev2'][0]['region'], 'South')
        self.assertEqual(
            public_ep['volumev2'][0]['url'],
            "http://cinder.south.host/cinderapi/public/v2",
        )
        self.assertEqual(public_ep['volumev3'][0]['region'], 'South')
        self.assertEqual(
            public_ep['volumev3'][0]['url'],
            "http://cinder.south.host/cinderapi/public/v3",
        )
        self.assertEqual(public_ep['block-storage'][0]['region'], 'North')
        self.assertEqual(
            public_ep['block-storage'][0]['url'],
            "http://cinder.north.host/cinderapi/public",
        )

    def test_service_catalog_regions(self):
        self.AUTH_RESPONSE_BODY['token']['region_name'] = "North"
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog

        url = sc.url_for(service_type='image', interface='public')
        self.assertEqual(url, "http://glance.north.host/glanceapi/public")

        self.AUTH_RESPONSE_BODY['token']['region_name'] = "South"
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        sc = auth_ref.service_catalog
        url = sc.url_for(
            service_type='image', region_name="South", interface='internal'
        )
        self.assertEqual(url, "http://glance.south.host/glanceapi/internal")

    def test_service_catalog_empty(self):
        self.AUTH_RESPONSE_BODY['token']['catalog'] = []
        auth_ref = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        )
        self.assertRaises(
            exceptions.EmptyCatalog,
            auth_ref.service_catalog.url_for,
            service_type='image',
            interface='internalURL',
        )

    def test_service_catalog_get_endpoints_region_names(self):
        sc = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        ).service_catalog

        endpoints = sc.get_endpoints(service_type='image', region_name='North')
        self.assertEqual(len(endpoints), 1)
        for endpoint in endpoints['image']:
            self.assertEqual(
                endpoint['url'], self.north_endpoints[endpoint['interface']]
            )

        endpoints = sc.get_endpoints(service_type='image', region_name='South')
        self.assertEqual(len(endpoints), 1)
        for endpoint in endpoints['image']:
            self.assertEqual(
                endpoint['url'], self.south_endpoints[endpoint['interface']]
            )

        endpoints = sc.get_endpoints(service_type='compute')
        self.assertEqual(len(endpoints['compute']), 3)

        endpoints = sc.get_endpoints(
            service_type='compute', region_name='North'
        )
        self.assertEqual(len(endpoints['compute']), 3)

        endpoints = sc.get_endpoints(
            service_type='compute', region_name='West'
        )
        self.assertEqual(len(endpoints['compute']), 0)

    def test_service_catalog_url_for_region_names(self):
        sc = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        ).service_catalog

        url = sc.url_for(service_type='image', region_name='North')
        self.assertEqual(url, self.north_endpoints['public'])

        url = sc.url_for(service_type='image', region_name='South')
        self.assertEqual(url, self.south_endpoints['public'])

        self.assertRaises(
            exceptions.EndpointNotFound,
            sc.url_for,
            service_type='image',
            region_name='West',
        )

    def test_service_catalog_get_url_region_names(self):
        sc = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        ).service_catalog

        urls = sc.get_urls(service_type='image')
        self.assertEqual(len(urls), 2)

        urls = sc.get_urls(service_type='image', region_name='North')
        self.assertEqual(len(urls), 1)
        self.assertEqual(urls[0], self.north_endpoints['public'])

        urls = sc.get_urls(service_type='image', region_name='South')
        self.assertEqual(len(urls), 1)
        self.assertEqual(urls[0], self.south_endpoints['public'])

        urls = sc.get_urls(service_type='image', region_name='West')
        self.assertEqual(len(urls), 0)

    def test_service_catalog_service_name(self):
        sc = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        ).service_catalog

        url = sc.url_for(
            service_name='glance',
            interface='public',
            service_type='image',
            region_name='North',
        )
        self.assertEqual('http://glance.north.host/glanceapi/public', url)

        url = sc.url_for(
            service_name='glance',
            interface='public',
            service_type='image',
            region_name='South',
        )
        self.assertEqual('http://glance.south.host/glanceapi/public', url)

        self.assertRaises(
            exceptions.EndpointNotFound,
            sc.url_for,
            service_name='glance',
            service_type='compute',
        )

        urls = sc.get_urls(
            service_type='image', service_name='glance', interface='public'
        )

        self.assertIn('http://glance.north.host/glanceapi/public', urls)
        self.assertIn('http://glance.south.host/glanceapi/public', urls)

        urls = sc.get_urls(
            service_type='image', service_name='Servers', interface='public'
        )

        self.assertEqual(0, len(urls))

    def test_service_catalog_without_name(self):
        f = fixture.V3Token(audit_chain_id=uuid.uuid4().hex)

        if not f.project_id:
            f.set_project_scope()

        f.add_role(name='admin')
        f.add_role(name='member')

        region = 'RegionOne'
        tenant = '225da22d3ce34b15877ea70b2a575f58'

        s = f.add_service('volume')
        s.add_standard_endpoints(
            public=f'http://public.com:8776/v1/{tenant}',
            internal=f'http://internal:8776/v1/{tenant}',
            admin=f'http://admin:8776/v1/{tenant}',
            region=region,
        )

        s = f.add_service('image')
        s.add_standard_endpoints(
            public='http://public.com:9292/v1',
            internal='http://internal:9292/v1',
            admin='http://admin:9292/v1',
            region=region,
        )

        s = f.add_service('compute')
        s.add_standard_endpoints(
            public=f'http://public.com:8774/v2/{tenant}',
            internal=f'http://internal:8774/v2/{tenant}',
            admin=f'http://admin:8774/v2/{tenant}',
            region=region,
        )

        s = f.add_service('ec2')
        s.add_standard_endpoints(
            public='http://public.com:8773/services/Cloud',
            internal='http://internal:8773/services/Cloud',
            admin='http://admin:8773/services/Admin',
            region=region,
        )

        s = f.add_service('identity')
        s.add_standard_endpoints(
            public='http://public.com:5000/v3',
            internal='http://internal:5000/v3',
            admin='http://admin:35357/v3',
            region=region,
        )

        pr_auth_ref = access.create(body=f)
        pr_sc = pr_auth_ref.service_catalog

        # this will work because there are no service names on that token
        url_ref = 'http://public.com:8774/v2/225da22d3ce34b15877ea70b2a575f58'
        url = pr_sc.url_for(
            service_type='compute', service_name='NotExist', interface='public'
        )
        self.assertEqual(url_ref, url)

        ab_auth_ref = access.create(body=self.AUTH_RESPONSE_BODY)
        ab_sc = ab_auth_ref.service_catalog

        # this won't work because there is a name and it's not this one
        self.assertRaises(
            exceptions.EndpointNotFound,
            ab_sc.url_for,
            service_type='compute',
            service_name='NotExist',
            interface='public',
        )


class ServiceCatalogV3Test(ServiceCatalogTest):
    def test_building_a_service_catalog(self):
        sc = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        ).service_catalog

        self.assertEqual(
            sc.url_for(service_type='compute'),
            'https://compute.north.host/novapi/public',
        )
        self.assertEqual(
            sc.url_for(service_type='compute', interface='internal'),
            'https://compute.north.host/novapi/internal',
        )

        self.assertRaises(
            exceptions.EndpointNotFound,
            sc.url_for,
            region_name='South',
            service_type='compute',
        )

    def test_service_catalog_endpoints(self):
        sc = access.create(
            auth_token=uuid.uuid4().hex, body=self.AUTH_RESPONSE_BODY
        ).service_catalog

        public_ep = sc.get_endpoints(
            service_type='compute', interface='public'
        )
        self.assertEqual(public_ep['compute'][0]['region_id'], 'North')
        self.assertEqual(
            public_ep['compute'][0]['url'],
            'https://compute.north.host/novapi/public',
        )

    def test_service_catalog_multiple_service_types(self):
        token = fixture.V3Token()
        token.set_project_scope()

        for i in range(3):
            s = token.add_service('compute')
            s.add_standard_endpoints(
                public=f'public-{i}',
                admin=f'admin-{i}',
                internal=f'internal-{i}',
                region=f'region-{i}',
            )

        auth_ref = access.create(resp=None, body=token)

        urls = auth_ref.service_catalog.get_urls(
            service_type='compute', interface='public'
        )

        self.assertEqual({'public-0', 'public-1', 'public-2'}, set(urls))

        urls = auth_ref.service_catalog.get_urls(
            service_type='compute', interface='public', region_name='region-1'
        )

        self.assertEqual(('public-1',), urls)

    def test_service_catalog_endpoint_id(self):
        token = fixture.V3Token()
        token.set_project_scope()

        service_id = uuid.uuid4().hex
        endpoint_id = uuid.uuid4().hex
        public_url = uuid.uuid4().hex

        s = token.add_service('compute', id=service_id)
        s.add_endpoint('public', public_url, id=endpoint_id)
        s.add_endpoint('public', uuid.uuid4().hex)

        auth_ref = access.create(body=token)

        # initially assert that we get back all our urls for a simple filter
        urls = auth_ref.service_catalog.get_urls(
            service_type='compute', interface='public'
        )
        self.assertEqual(2, len(urls))

        # with bad endpoint_id nothing should be found
        urls = auth_ref.service_catalog.get_urls(
            service_type='compute',
            endpoint_id=uuid.uuid4().hex,
            interface='public',
        )

        self.assertEqual(0, len(urls))

        # with service_id we get back both public endpoints
        urls = auth_ref.service_catalog.get_urls(
            service_type='compute', service_id=service_id, interface='public'
        )
        self.assertEqual(2, len(urls))

        # with service_id and endpoint_id we get back the url we want
        urls = auth_ref.service_catalog.get_urls(
            service_type='compute',
            service_id=service_id,
            endpoint_id=endpoint_id,
            interface='public',
        )

        self.assertEqual((public_url,), urls)

        # with service_id and endpoint_id we get back the url we want
        urls = auth_ref.service_catalog.get_urls(
            service_type='compute', endpoint_id=endpoint_id, interface='public'
        )

        self.assertEqual((public_url,), urls)

    def test_service_catalog_without_service_type(self):
        token = fixture.V3Token()
        token.set_project_scope()

        public_urls = []

        for i in range(0, 3):
            public_url = uuid.uuid4().hex
            public_urls.append(public_url)

            s = token.add_service(uuid.uuid4().hex)
            s.add_endpoint('public', public_url)

        auth_ref = access.create(body=token)
        urls = auth_ref.service_catalog.get_urls(interface='public')

        self.assertEqual(3, len(urls))

        for p in public_urls:
            self.assertIn(p, urls)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import exceptions
from keystoneauth1 import identity
from keystoneauth1.identity.v3 import oidc
from keystoneauth1 import loading
from keystoneauth1.loading import opts


def _add_common_identity_options(options: list[opts.Opt]) -> None:
    options.extend(
        [
            loading.Opt('user-id', help="User's user ID"),
            loading.Opt(
                'username',
                help="User's username",
                deprecated=[loading.Opt('user-name')],
            ),
            loading.Opt('user-domain-id', help="User's domain ID"),
            loading.Opt('user-domain-name', help="User's domain name"),
        ]
    )


def _assert_identity_options(options: dict[str, ty.Any]) -> None:
    if options.get('username') and not (
        options.get('user_domain_name') or options.get('user_domain_id')
    ):
        m = (
            "You have provided a username. In the V3 identity API a "
            "username is only unique within a domain so you must "
            "also provide either a user_domain_id or user_domain_name."
        )
        raise exceptions.OptionError(m)


class Password(loading.BaseV3Loader[identity.V3Password]):
    """Authenticate with a username and password.

    Authenticate to the identity service using the provided username and
    password. This is the standard and most common form of authentication.
    """

    @property
    def plugin_class(self) -> type[identity.V3Password]:
        return identity.V3Password

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()
        _add_common_identity_options(options)

        options.extend(
            [
                loading.Opt(
                    'password',
                    secret=True,
                    prompt='Password: ',
                    help="User's password",
                )
            ]
        )

        return options

    def load_from_options(self, **kwargs: ty.Any) -> identity.V3Password:
        _assert_identity_options(kwargs)

        return super().load_from_options(**kwargs)


class Token(loading.BaseV3Loader[identity.V3Token]):
    """Given an existing token rescope it to another target.

    Use the Identity service's rescope mechanism to get a new token based upon
    an existing token. Because an auth plugin requires a service catalog and
    scope information it is often easier to fetch a new token based on an
    existing one than validate and reuse the one you already have.
    """

    @property
    def plugin_class(self) -> type[identity.V3Token]:
        return identity.V3Token

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'token', secret=True, help='Token to authenticate with'
                )
            ]
        )

        return options


class _OpenIDConnectBase(loading.BaseFederationLoader[oidc._OidcBaseT]):
    def load_from_options(self, **kwargs: ty.Any) -> oidc._OidcBaseT:
        if not (
            kwargs.get('access_token_endpoint')
            or kwargs.get('discovery_endpoint')
        ):
            m = (
                "You have to specify either an 'access-token-endpoint' or "
                "a 'discovery-endpoint'."
            )
            raise exceptions.OptionError(m)

        return super().load_from_options(**kwargs)

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt('client-id', help='OAuth 2.0 Client ID'),
                loading.Opt(
                    'client-secret',
                    secret=True,
                    help='OAuth 2.0 Client Secret',
                ),
                loading.Opt(
                    'openid-scope',
                    default="openid profile",
                    dest="scope",
                    help='OpenID Connect scope that is requested from '
                    'authorization server. Note that the OpenID '
                    'Connect specification states that "openid" '
                    'must be always specified.',
                ),
                loading.Opt(
                    'access-token-endpoint',
                    help='OpenID Connect Provider Token Endpoint. Note '
                    'that if a discovery document is being passed this '
                    'option will override the endpoint provided by the '
                    'server in the discovery document.',
                ),
                loading.Opt(
                    'discovery-endpoint',
                    help='OpenID Connect Discovery Document URL. '
                    'The discovery document will be used to obtain the '
                    'values of the access token endpoint and the '
                    'authentication endpoint. This URL should look like '
                    'https://idp.example.org/.well-known/'
                    'openid-configuration',
                ),
                loading.Opt(
                    'access-token-type',
                    help='OAuth 2.0 Authorization Server Introspection '
                    'token type, it is used to decide which type '
                    'of token will be used when processing token '
                    'introspection. Valid values are: '
                    '"access_token" or "id_token"',
                ),
            ]
        )

        return options


class OpenIDConnectClientCredentials(
    _OpenIDConnectBase[identity.V3OidcClientCredentials]
):
    """Authenticate with the OIDC Client Credentials flow."""

    @property
    def plugin_class(self) -> type[identity.V3OidcClientCredentials]:
        return identity.V3OidcClientCredentials

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        return options


class OpenIDConnectPassword(_OpenIDConnectBase[identity.V3OidcPassword]):
    """Authenticate with the OIDC Resource Owner Password Credentials flow."""

    @property
    def plugin_class(self) -> type[identity.V3OidcPassword]:
        return identity.V3OidcPassword

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt('username', help='Username', required=True),
                loading.Opt(
                    'password', secret=True, help='Password', required=True
                ),
                loading.Opt(
                    'idp_otp_key',
                    help='A key to be used in the Identity Provider access'
                    ' token endpoint to pass the OTP value. '
                    'E.g. totp',
                ),
            ]
        )

        return options


class OpenIDConnectAuthorizationCode(
    _OpenIDConnectBase[identity.V3OidcAuthorizationCode]
):
    """Authenticate with the OIDC Authorization Code flow."""

    @property
    def plugin_class(self) -> type[identity.V3OidcAuthorizationCode]:
        return identity.V3OidcAuthorizationCode

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'redirect-uri', help='OpenID Connect Redirect URL'
                ),
                loading.Opt(
                    'code',
                    secret=True,
                    required=True,
                    deprecated=[loading.Opt('authorization-code')],
                    help='OAuth 2.0 Authorization Code',
                ),
            ]
        )

        return options


class OpenIDConnectAccessToken(
    loading.BaseFederationLoader[identity.V3OidcAccessToken]
):
    """Authenticate with the OIDC Access Token flow."""

    @property
    def plugin_class(self) -> type[identity.V3OidcAccessToken]:
        return identity.V3OidcAccessToken

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'access-token',
                    secret=True,
                    required=True,
                    help='OAuth 2.0 Access Token',
                )
            ]
        )
        return options


class OpenIDConnectDeviceAuthorization(
    _OpenIDConnectBase[identity.V3OidcDeviceAuthorization]
):
    """Authenticate with the OAuth 2.0 Device Authorization flow."""

    @property
    def plugin_class(self) -> type[identity.V3OidcDeviceAuthorization]:
        return identity.V3OidcDeviceAuthorization

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'device-authorization-endpoint',
                    help='OAuth 2.0 Device Authorization Endpoint. Note '
                    'that if a discovery document is being passed this '
                    'option will override the endpoint provided by the '
                    'server in the discovery document.',
                ),
                loading.Opt(
                    'code-challenge-method',
                    help='PKCE Challenge Method (RFC 7636)',
                ),
            ]
        )

        return options


class TOTP(loading.BaseV3Loader[identity.V3TOTP]):
    """Authenticate with a Time-based One-Time Password.

    Authenticate to the identity service using a time-based one-time password.
    This is typically used in combination with another plugin as part of a
    multi-factor configuration.
    """

    @property
    def plugin_class(self) -> type[identity.V3TOTP]:
        return identity.V3TOTP

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()
        _add_common_identity_options(options)

        options.extend(
            [
                loading.Opt(
                    'passcode',
                    secret=True,
                    prompt='TOTP passcode: ',
                    help="User's TOTP passcode",
                )
            ]
        )

        return options

    def load_from_options(self, **kwargs: ty.Any) -> identity.V3TOTP:
        _assert_identity_options(kwargs)

        return super().load_from_options(**kwargs)


class TokenlessAuth(loading.BaseLoader[identity.V3TokenlessAuth]):
    """Authenticate without a token, using an X.509 certificate."""

    @property
    def plugin_class(self) -> type[identity.V3TokenlessAuth]:
        return identity.V3TokenlessAuth

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'auth-url', required=True, help='Authentication URL'
                ),
                loading.Opt('domain-id', help='Domain ID to scope to'),
                loading.Opt('domain-name', help='Domain name to scope to'),
                loading.Opt('project-id', help='Project ID to scope to'),
                loading.Opt('project-name', help='Project name to scope to'),
                loading.Opt(
                    'project-domain-id', help='Domain ID containing project'
                ),
                loading.Opt(
                    'project-domain-name',
                    help='Domain name containing project',
                ),
            ]
        )

        return options

    def load_from_options(self, **kwargs: ty.Any) -> identity.V3TokenlessAuth:
        if (
            not kwargs.get('domain_id')
            and not kwargs.get('domain_name')
            and not kwargs.get('project_id')
            and not kwargs.get('project_name')
            or (
                kwargs.get('project_name')
                and not (
                    kwargs.get('project_domain_name')
                    or kwargs.get('project_domain_id')
                )
            )
        ):
            m = (
                'You need to provide either a domain_name, domain_id, '
                'project_id or project_name. '
                'If you have provided a project_name, in the V3 identity '
                'API a project_name is only unique within a domain so '
                'you must also provide either a project_domain_id or '
                'project_domain_name.'
            )
            raise exceptions.OptionError(m)

        return super().load_from_options(**kwargs)


class ApplicationCredential(
    loading.BaseV3Loader[identity.V3ApplicationCredential]
):
    """Authenticate with an application credential.

    Authenticate to the identity service using the provided application
    credential secret and ID or name. If a name is used, you must also provide
    a username and user domain to assist in lookup.
    """

    @property
    def plugin_class(self) -> type[identity.V3ApplicationCredential]:
        return identity.V3ApplicationCredential

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()
        _add_common_identity_options(options)

        options.extend(
            [
                loading.Opt(
                    'application_credential_secret',
                    secret=True,
                    required=True,
                    help="Application credential auth secret",
                ),
                loading.Opt(
                    'application_credential_id',
                    help='Application credential ID',
                ),
                loading.Opt(
                    'application_credential_name',
                    help='Application credential name',
                ),
            ]
        )

        return options

    def load_from_options(
        self, **kwargs: ty.Any
    ) -> identity.V3ApplicationCredential:
        _assert_identity_options(kwargs)
        if not kwargs.get('application_credential_id') and not kwargs.get(
            'application_credential_name'
        ):
            m = (
                'You must provide either an application credential ID or an '
                'application credential name and user.'
            )
            raise exceptions.OptionError(m)
        if not kwargs.get('application_credential_secret'):
            m = 'You must provide an auth secret.'
            raise exceptions.OptionError(m)

        return super().load_from_options(**kwargs)


class MultiFactor(loading.BaseV3Loader[identity.V3MultiFactor]):
    """Authenticate using multiple factors.

    Authenticate to the identity service using a combination of factors, such
    as username/password and a TOTP code.
    """

    def __init__(self) -> None:
        super().__init__()
        self._methods = None

    @property
    def plugin_class(self) -> type[identity.V3MultiFactor]:
        return identity.V3MultiFactor

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'auth_methods',
                    required=True,
                    help="Methods to authenticate with.",
                )
            ]
        )

        if self._methods:
            options_dict = {o.name: o for o in options}
            for method in self._methods:
                method_opts = loading.get_plugin_options(method)
                for opt in method_opts:
                    options_dict[opt.name] = opt
            options = list(options_dict.values())
        return options

    def load_from_options(self, **kwargs: ty.Any) -> identity.V3MultiFactor:
        _assert_identity_options(kwargs)

        if 'auth_methods' not in kwargs:
            raise exceptions.OptionError("methods is a required option.")

        self._methods = kwargs['auth_methods']

        return super().load_from_options(**kwargs)


class OAuth2ClientCredential(
    loading.BaseV3Loader[identity.V3OAuth2ClientCredential]
):
    """Authenticate with an OAuth2.0 client credential."""

    @property
    def plugin_class(self) -> type[identity.V3OAuth2ClientCredential]:
        return identity.V3OAuth2ClientCredential

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()
        options.extend(
            [
                loading.Opt(
                    'oauth2_endpoint',
                    required=True,
                    help='Endpoint for OAuth2.0',
                ),
                loading.Opt(
                    'oauth2_client_id',
                    required=True,
                    help='Client id for OAuth2.0',
                ),
                loading.Opt(
                    'oauth2_client_secret',
                    secret=True,
                    required=True,
                    help='Client secret for OAuth2.0',
                ),
            ]
        )

        return options

    def load_from_options(
        self, **kwargs: ty.Any
    ) -> identity.V3OAuth2ClientCredential:
        if not kwargs.get('oauth2_endpoint'):
            m = 'You must provide an OAuth2.0 endpoint.'
            raise exceptions.OptionError(m)
        if not kwargs.get('oauth2_client_id'):
            m = 'You must provide an OAuth2.0 client credential ID.'
            raise exceptions.OptionError(m)
        if not kwargs.get('oauth2_client_secret'):
            m = 'You must provide an OAuth2.0 client credential auth secret.'
            raise exceptions.OptionError(m)

        return super().load_from_options(**kwargs)


class OAuth2mTlsClientCredential(
    loading.BaseV3Loader[identity.V3OAuth2mTlsClientCredential]
):
    """Authenticate with an OAuth2.0 mTLS client credential."""

    @property
    def plugin_class(self) -> type[identity.V3OAuth2mTlsClientCredential]:
        return identity.V3OAuth2mTlsClientCredential

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()
        options.extend(
            [
                loading.Opt(
                    'oauth2-endpoint',
                    required=True,
                    help='Endpoint for OAuth2.0 Mutual-TLS Authorization',
                ),
                loading.Opt(
                    'oauth2-client-id',
                    required=True,
                    help='Client credential ID for OAuth2.0 Mutual-TLS '
                    'Authorization',
                ),
            ]
        )
        return options

    def load_from_options(
        self, **kwargs: ty.Any
    ) -> identity.V3OAuth2mTlsClientCredential:
        if not kwargs.get('oauth2_endpoint'):
            m = 'You must provide an OAuth2.0 Mutual-TLS endpoint.'
            raise exceptions.OptionError(m)
        if not kwargs.get('oauth2_client_id'):
            m = (
                'You must provide an client credential ID for '
                'OAuth2.0 Mutual-TLS Authorization.'
            )
            raise exceptions.OptionError(m)
        return super().load_from_options(**kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os

import testtools
import yaml

from keystoneauth1.fixture import serializer


class TestBetamaxSerializer(testtools.TestCase):
    TEST_FILE = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'data',
        'ksa_betamax_test_cassette.yaml',
    )
    TEST_JSON = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'data',
        'ksa_serializer_data.json',
    )

    def setUp(self):
        super().setUp()
        self.serializer = serializer.YamlJsonSerializer()

    def test_deserialize(self):
        data = self.serializer.deserialize(open(self.TEST_FILE).read())
        request = data['http_interactions'][0]['request']
        self.assertEqual(
            'http://keystoneauth-betamax.test/v2.0/tokens', request['uri']
        )
        payload = json.loads(request['body']['string'])
        self.assertEqual('test_tenant_name', payload['auth']['tenantName'])

    def test_serialize(self):
        data = json.loads(open(self.TEST_JSON).read())
        serialized = self.serializer.serialize(data)
        data = yaml.safe_load(serialized)
        request = data['http_interactions'][0]['request']
        self.assertEqual(
            'http://keystoneauth-betamax.test/v2.0/tokens', request['uri']
        )
        payload = json.loads(request['body']['string'])
        self.assertEqual('test_tenant_name', payload['auth']['tenantName'])
